from typing import Iterable

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .select import Select
from .col import Col


@typechecked
class OrderBy(ChocExpr):
    """ OrderBy expression """
    def __init__(
            self,
            select: Select,
            compact: bool = True
    ) -> None:
        self._select = select
        super().__init__("ORDER BY {$(cols).ordering_label}", compact=compact)

    @property
    def cols(self) -> Iterable[Col]:
        return [col for col in self._select.cols if col.ordering_label]

    @property
    def buildable(self) -> bool:
        return len(self.cols) > 0



from choc_expr import Expr as ChocExpr

from ..expr.query import Query
from ..view_mode import ViewMode


class View(ChocExpr):

    def __init__(self, name: str, query: Query | None = None, mode=ViewMode.Create, compact: bool = True) -> None:
        self._name = name
        self._query = query
        self._mode = mode
        super().__init__("@{create_mode}:CREATE VIEW {_name} AS~{_query}:;@{drop_mode}:DROP VIEW {_name}~:;", compact=compact)

    @property
    def create_mode(self):
        return self._mode == ViewMode.Create

    @property
    def drop_mode(self):
        return self._mode == ViewMode.Drop



from __future__ import annotations
from typing import Any, Self, TYPE_CHECKING

if TYPE_CHECKING:
    from .expr.col import Col


def quote_expr(expr: int | float | str | Col | Self) -> str:
    """ Add apostrophes arround a string if it is one, otherwise do nothing """
    if type(expr) is str:
        return f"'{expr}'"
    return expr


def to_bool(val: Any) -> bool:
    """ Cast any value into boolean """
    if not val or val in ("False", "None"):
        return False
    return True



from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .table import Table


@typechecked
class Update(ChocExpr):
    """ Update expression """
    def __init__(
            self,
            table: Table | str | None = None,
            compact: bool = True
    ) -> None:
        self.table = table
        super().__init__("UPDATE {table}", compact=compact)

    @property
    def table(self):
        return self._table

    @table.setter
    def table(self, value):
        self._table = Table(value) if type(value) is str else value

    @property
    def buildable(self) -> bool:
        return self.table is not None



from enum import Enum


class SqlFunction(Enum):

    Length = "LENGTH"
    Replace = "REPLACE"
    Soundex = "SOUNDEX"
    Substring = "SUBSTRING"
    Reverse = "REVERSE"
    Trim = "TRIM"
    LTrim = "LTRIM"
    RTrim = "RTRIM"
    LPad = "LPAD"
    Locate = "LOCATE"
    Instr = "INSTR"
    Rand = "RAND"
    DateFormat = "DATE_FORMAT"
    DayOfWeek = "DAYOFWEEK"
    SecToTime = "SEC_TO_TIME"
    TimeDiff = "TIMEDIFF"
    TimeStamp = "TIMESTAMP"
    MD5 = "MD5"
    Cast = "CAST"
    Convert = "CONVERT"
    IsNull = "ISNULL"
    GroupConcat = "GROUP_CONCAT"
    Upper = "UPPER"
    Lower = "LOWER"
    Left = "LEFT"
    Right = "RIGHT"
    Abs = "ABS"
    Round = "ROUND"
    Ceiling = "CEILING"
    Floor = "FLOOR"
    Second = "SECOND"
    Minute = "MINUTE"
    Hour = "HOUR"
    Day = "DAY"
    Month = "MONTH"
    Year = "YEAR"
    Now = "NOW"
    Date = "DATE"
    DateDiff = "DATEDIFF"
    Coalesce = "COALESCE"



from typing import Iterable
from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .col import Col
from .condition import Condition


@typechecked
class Set(ChocExpr):
    """ Set expression """
    def __init__(
            self,
            assignations: Iterable[Condition] | None = None,
            compact: bool = True
    ) -> None:
        self.assignations = assignations or []
        super().__init__("SET {$(assignations)}", compact=compact)

    @property
    def assignations(self):
        return self._cols

    @assignations.setter
    def assignations(self, value):
        self._cols = [Col(col) if type(col) is str else col for col in (value or [])]

    @property
    def buildable(self):
        return len(self.assignations) > 0



from enum import Enum


class Operator(Enum):

    LowerThan = "<"
    GreaterThan = ">"
    LowerOrEqualThan = "<="
    GreaterOrEqualThan = ">="
    Equal = "="
    NotEqual = "<>"
    And = "AND"
    Or = "OR"
    Like = "LIKE"
    In = "IN"



from typeguard import typechecked

from .agg_function import AggFunction
from .sql_function import SqlFunction
from .expr.col import Col
from .ordering import Ordering


@typechecked
def concat(*cols: Col) -> Col:
    """ Shortcut to concatenate some columns """
    if len(cols) < 2:
        raise ValueError("You cannot concatenate less than two cols")
    res = None
    for k, col in enumerate(cols):
        res = col if k == 0 else res & col
    return res


@typechecked
def count(col_name: str | None = None) -> Col:
    """ Shortcut to create a column and apply the "count" aggregation function """
    return Col(name=col_name or "*", agg_function=AggFunction.Count)


@typechecked
def sum(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "sum" aggregation function """
    return Col(name=col_name, agg_function=AggFunction.Sum)


@typechecked
def upper(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "upper" SQL function """
    return Col(name=col_name, sql_function=SqlFunction.Upper)


@typechecked
def lower(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "upper" SQL function """
    return Col(name=col_name, sql_function=SqlFunction.Lower)


@typechecked
def second(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "second" SQL function """
    return Col(name=col_name, sql_function=SqlFunction.Second)


@typechecked
def minute(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "minute" SQL function """
    return Col(name=col_name, sql_function=SqlFunction.Minute)


@typechecked
def hour(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "hour" SQL function """
    return Col(name=col_name, sql_function=SqlFunction.Hour)


@typechecked
def day(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "day" SQL function """
    return Col(name=col_name, sql_function=SqlFunction.Day)


@typechecked
def month(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "month" SQL function """
    return Col(name=col_name, sql_function=SqlFunction.Month)


@typechecked
def year(col_name: str) -> Col:
    """ Shortcut to create a column and apply the "year" SQL function """
    return Col(name=col_name, sql_function=SqlFunction.Year)


@typechecked
def asc(col_name: str) -> Col:
    """ Shortcut to create a column and apply the ascending ordering """
    return Col(name=col_name, ordering=Ordering.Ascending)


@typechecked
def desc(col_name: str) -> Col:
    """ Shortcut to create a column and apply the descending ordering  """
    return Col(name=col_name, ordering=Ordering.Descending)



from __future__ import annotations
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from .table import Table
    from .condition import Condition

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .update import Update
from .set import Set


@typechecked
class UpdateSet(ChocExpr):
    """ UpdateSet expression """
    def __init__(
            self,
            table: str | Table | None = None,
            assignations: Iterable[Condition] | None = None,
            compact: bool = True
    ) -> None:
        self._update = Update(table=table, compact=compact)
        self._set = Set(assignations=assignations, compact=compact)
        super().__init__("{_update~}{_set}", compact=compact)

    @property
    def buildable(self) -> bool:
        return self._update.buildable and self._set.buildable



from __future__ import annotations

import re
from typing import Any, Iterable, Self, TYPE_CHECKING



if TYPE_CHECKING:
    from .query import Query

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .when import When
from .condition import Condition
from ..sql_type import SqlType
from ..operator import Operator
from ..ordering import Ordering
from ..agg_function import AggFunction
from ..sql_function import SqlFunction


@typechecked
class Col(ChocExpr):
    """ SQL column """
    def __init__(
            self,
            name: str,
            alias: str | None = None,
            agg_function: AggFunction | None = None,
            sql_function: SqlFunction | None = None,
            ordering: Ordering | None = None,
            table_name: str | None = None,
            type: SqlType | None = None
    ) -> None:
        if not name:
            raise ValueError("The name parameter must not be empty")
        if name.startswith("<:"):
            name = name[2:]
            ordering = Ordering.Descending
        elif name.startswith(">:"):
            name = name[2:]
            ordering = Ordering.Ascending
        if name != "*":
            match = re.search(r"^([A-Za-z_\s]+\.)?([A-Za-z_\s()]+){1}((?:@|:)[A-Za-z_\s]*)?$", name)
            if not match:
                raise ValueError(f"Forbidden characters in expr '{name}'")
            r, n, a = match.groups()
            name = n
            alias = alias or (a[1:] if a else None)
            table_name = table_name or (r[:-1] if r else None)
        else:
            if alias:
                raise ValueError("alias parameter can't be used with star expression")
            if table_name:
                raise ValueError("table_name parameter can't be used with star expression")
            if ordering:
                raise ValueError("ordering parameter can't be used with star expression")

        self._agg_function = agg_function
        self._sql_function = sql_function
        self._ordering = ordering
        self._concatenation = []
        self._alias = alias
        self._table_name = table_name
        self._name = name
        self._type = type

        super().__init__("{full_name}@{_alias}: AS {_alias}:;")

    @property
    def full_name(self):
        if self._concatenation:
            name = f"CONCAT({", ".join(list(f"'{x}'" if type(x) is str else str(x) for x in self._concatenation))})"
        else:
            name = f"{self._table_name + "." if self._table_name else ""}{self._name}"
        if self._agg_function:
            return f"{self._agg_function.value}({f"{name}"})"
        if self._sql_function:
            return f"{self._sql_function.value}({f"{name}"})"
        return f"{name}"
    
    @property
    def creation_name(self):
        return f"{self._name} {self._type.value}"

    def copy(self) -> Self:
        """ Copy the column """
        return Col(
            name=self._name,
            alias=self._alias,
            agg_function=self._agg_function,
            sql_function=self._sql_function,
            ordering=self._ordering,
            table_name=self._table_name
        )

    def __gt__(self, value: Col | int | float | When) -> Condition:
        return Condition(left_value=self, op=Operator.GreaterThan, right_value=value)

    def __eq__(self, value: Col | int | float | str | When) -> Condition:
        return Condition(left_value=self, op=Operator.Equal, right_value=value)

    def __ne__(self, value: Col | int | float | str | When) -> Condition:
        return Condition(left_value=self, op=Operator.NotEqual, right_value=value)

    def __and__(self, value: Col | str) -> Self:
        if not self._concatenation:
            self._concatenation.append(self.copy())
        self._sql_function = None
        self._concatenation.append(value)
        return self

    def __rand__(self, value: Col | str) -> Self:
        return self.__and__(value)

    def __rshift__(self, value: str) -> Condition:
        return self.like(value)

    def __lshift__(self, expr: Query | Iterable[int | float | str]) -> Condition:
        return self.isin(expr)

    def order(self, ordering: Ordering = Ordering.Ascending) -> Self:
        """ Order a column """
        self._ordering = ordering
        return self

    def asc(self) -> Self:
        """ Order a column by ascending order """
        return self.order(Ordering.Ascending)

    def desc(self) -> Self:
        """ Order a column by descending order """
        return self.order(Ordering.Descending)

    def like(self, expr: str) -> Condition:
        """ Apply the "like" operator """
        return Condition(self, Operator.Like, expr)

    def isin(self, expr: Query | Iterable[int | float | str]) -> Condition:
        """ Apply the "in" operator """
        return Condition(self, Operator.In, expr)

    def upper(self) -> Self:
        """ Apply the "upper" operator """
        self._sql_function = SqlFunction.Upper
        return self

    def lower(self) -> Self:
        """ Apply the "lower" operator """
        self._sql_function = SqlFunction.Lower
        return self

    def alias(self, name: str) -> Self:
        """ Set an alias """
        if name.startswith("<:"):
            name = name[2:]
            self._ordering = Ordering.Descending
        elif name.startswith(">:"):
            name = name[2:]
            self._ordering = Ordering.Ascending
        self._alias = name
        return self

    def remove_alias(self) -> Self:
        """ Remove the alias """
        self._alias = None
        return self

    def aggregate(self, agg_function: AggFunction) -> Self:
        """ Set an aggregate function """
        self._agg_function = agg_function
        return self

    def sum(self) -> Self:
        """ Apply the "sum" function """
        self._agg_function = AggFunction.Sum
        return self

    def count(self) -> Self:
        """ Apply the "count" function """
        self._agg_function = AggFunction.Count
        return self

    def max(self) -> Self:
        """ Apply the "max" function """
        self._agg_function = AggFunction.Max
        return self

    def min(self) -> Self:
        """ Apply the "min" function """
        self._agg_function = AggFunction.Min
        return self

    def average(self) -> Self:
        """ Apply the "avg" function """
        self._agg_function = AggFunction.Average
        return self

    @property
    def ordering_label(self) -> str | None:
        return f"{(self._table_name + ".") if self._table_name else ""}{self._alias if self._alias else self._name} {self._ordering.value}" if self._ordering else None



from typeguard import typechecked
from choc_expr import Expr as ChocExpr


@typechecked
class Limit(ChocExpr):
    """ Limit expression """
    def __init__(
            self,
            length: int | None,
            compact: bool = True
    ) -> None:
        if length is not None and length < 1:
            raise ValueError("Length cannot be lower than 1")
        self._length = length
        super().__init__("LIMIT {length}", compact=compact)

    @property
    def length(self) -> int:
        return self._length

    @property
    def buildable(self) -> bool:
        return self._length is not None and self._length > 0



from typing import Any, Iterable, Self, Tuple

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .delete_from import DeleteFrom
from ..query_mode import QueryMode
from ..alter_mode import AlterMode
from .join import Join
from .order_by import OrderBy
from .group_by import GroupBy
from .having import Having
from .where import Where
from .select_from import SelectFrom
from .limit import Limit
from ..agg_function import AggFunction
from ..join_type import JoinType
from .condition import Condition
from .col import Col
from .table import Table
from .update_set import UpdateSet
from .union import Union


@typechecked
class Query(ChocExpr):
    """ Handler to generate a SQL query """

    def __init__(
            self,
            query_mode: QueryMode = QueryMode.Select,
            compact: bool = True,
            limit: int | None = None,
            table: str | Table | None = None,
            unique: bool = False,
            joins: Iterable[Tuple[str | Table, Condition | str | Iterable[str]] | Tuple[str | Table, Condition | str | Iterable[str], JoinType | None]] | None = None,
            cols: Iterable[str | Col] | None = None,
            groups: Iterable[str] | None = None,
            filters: Iterable[Condition] | None = None,
            assignations: Iterable[Condition] | None = None,
            values: Iterable[Iterable[Any]] | None = None,
            auto_id: bool = False,
            alter_mode: AlterMode | None = None,
            alter_col: Col | str | None = None,
            alter_new_col: Col | str | None = None,
            alter_col_after: Col | str | None = None
    ) -> None:
        """"""
        self._compact = compact
        self._query_mode = query_mode
        self._table = table
        match query_mode:
            case QueryMode.Create:
                self._auto_id = auto_id
                if cols is None:
                    cols = []
                self._cols = cols
                super().__init__(
                    "CREATE TABLE {_table}~(@{_auto_id}:id MEDIUMINT NOT NULL AUTO_INCREMENT, :;{$(_cols).creation_name}@{_auto_id}:, PRIMARY KEY (id):;)",
                    compact=compact
                )
            case QueryMode.Insert:
                if cols is None:
                    cols = []
                self._cols = cols
                if values is None:
                    values = []
                self._values = values
                super().__init__(
                    "INSERT INTO {_table} ({$(_cols)})~VALUES ({$(_values)})",
                    compact=compact
                )
            case QueryMode.Alter:
                self._alter_mode = alter_mode
                if isinstance(alter_col, str):
                    alter_col = Col(alter_col)
                if isinstance(alter_col_after, str):
                    alter_col_after = Col(alter_col_after)
                if isinstance(alter_new_col, str):
                    alter_new_col = Col(alter_new_col)
                self._alter_col = alter_col
                self._alter_col_after = alter_col_after
                self._alter_new_col = alter_new_col
                match alter_mode:
                    case AlterMode.Add:
                        super().__init__(
                            "ALTER TABLE {_table}~ADD COLUMN {_alter_new_col.creation_name}@{_alter_col_after}: AFTER {_alter_col_after}:;~",
                            compact=compact
                        )
                    case AlterMode.Drop:
                        super().__init__(
                            "ALTER TABLE {_table}~DROP COLUMN {_alter_col}~",
                            compact=compact
                        )
                    case AlterMode.Rename:
                        super().__init__(
                            "ALTER TABLE {_table}~RENAME COLUMN {_alter_col} {_alter_new_col}~",
                            compact=compact
                        )
                    case AlterMode.Change:
                        super().__init__(
                            "ALTER TABLE {_table}~CHANGE COLUMN {_alter_col} {_alter_new_col.creation_name}~",
                            compact=compact
                        )
            case QueryMode.Delete:
                self._delete_from = DeleteFrom(table=table, compact=compact)
                if filters is None:
                    filters = []
                for filter in filters:
                    self.filter(filter)
                self._where = Where(compact=False)
                super().__init__(
                    "{_delete_from~}{_where~}",
                    list_join_sep="\n",
                    compact=compact
                )
            case QueryMode.Update:
                self._update_set = UpdateSet(table=table, assignations=assignations or [], compact=compact)
                self._where = Where(compact=False)
                if filters is None:
                    filters = []
                for filter in filters:
                    self.filter(filter)
                super().__init__(
                    "{_update_set~}{_where~}",
                    list_join_sep="\n",
                    compact=compact
                )
            case QueryMode.Select:
                if cols is None:
                    cols = []
                self._select_from = SelectFrom(table=table, unique=unique, compact=False)
                self.select(*cols)
                self._joins = []
                if joins is None:
                    joins = []
                if joins:
                    self.join_many(*joins)
                self._order_by = OrderBy(select=self._select_from.select, compact=False)
                if groups is None:
                    groups = []
                self._group_by = GroupBy(compact=False)
                if groups:
                    self.group_by(*groups)
                self._limit = Limit(length=limit, compact=False)
                self._where = Where(compact=False)
                if filters is None:
                    filters = []
                self._having = Having(compact=False)
                for filter in filters:
                    self.filter(filter)
                super().__init__(
                    "{_select_from~}{$(_joins)~}{_where~}{_group_by~}{_having~}{_order_by~}{_limit~}",
                    list_join_sep="\n",
                    compact=compact
                )

    @staticmethod
    def alter_table(
        table: str | Table,
        alter_mode: AlterMode,
        col: Col | str | None = None,
        new_col: Col | str | None = None,
        after: Col | str | None = None,
        compact: bool = True
    ):
        """"""
        return Query(query_mode=QueryMode.Alter, alter_mode=alter_mode, table=table, alter_col=col, alter_new_col=new_col, alter_col_after=after, compact=compact)
    
    @staticmethod
    def add_col(
        table: str | Table,
        col: Col,
        after: Col | str | None = None,
        compact: bool = True
    ):
        """"""
        return Query.alter_table(table=table, alter_mode=AlterMode.Add, new_col=col, after=after, compact=compact)
    
    @staticmethod
    def rename_col(
        table: str | Table,
        col: Col | str,
        new_col: Col | str,
        compact: bool = True
    ):
        """"""
        return Query.alter_table(table=table, alter_mode=AlterMode.Rename, col=col, new_col=new_col, compact=compact)
    
    @staticmethod
    def change_col(
        table: str | Table,
        col: Col | str,
        new_col: Col | str,
        compact: bool = True
    ):
        """"""
        return Query.alter_table(table=table, alter_mode=AlterMode.Change, col=col, new_col=new_col, compact=compact)
    
    @staticmethod
    def drop_col(
        table: str | Table,
        col: Col | str,
        compact: bool = True
    ):
        """"""
        return Query.alter_table(table=table, alter_mode=AlterMode.Drop, col=col, compact=compact)
    
    @staticmethod
    def insert_row(
        table: str | Table,
        cols: Iterable[Col],
        row: Iterable[Any],
        compact: bool = True
    ):
        """"""
        return Query(query_mode=QueryMode.Insert, table=table, cols=cols, values=(row,), compact=compact)
    
    @staticmethod
    def insert_rows(table: str | Table, cols: Iterable[Col], rows: Iterable[Iterable[Any]], compact: bool = True):
        """"""
        return Query(query_mode=QueryMode.Insert, table=table, cols=cols, values=rows, compact=compact)
    
    @staticmethod
    def delete_rows(table: str | Table, filters: Iterable[Condition], compact: bool = True):
        """"""
        return Query(query_mode=QueryMode.Delete, table=table, filters=filters, compact=compact)
    
    @staticmethod
    def update_rows(table: str | Table, filters: Iterable[Condition], assignations: Iterable[Condition], compact: bool = True):
        """"""
        return Query(query_mode=QueryMode.Delete, table=table, filters=filters, assignations=assignations, compact=compact)
    
    @staticmethod
    def create_table(table: str | Table, cols: Iterable[Col], auto_id: bool = False, compact: bool = True):
        """"""
        return Query(query_mode=QueryMode.Create, table=table, cols=cols, auto_id=auto_id, compact=compact)

    def compact(self):
        """"""
        self._compact = True

    @property
    def read_mode(self) -> bool:
        return self._query_mode == QueryMode.Select

    @property
    def update_mode(self) -> bool:
        return self._query_mode == QueryMode.Update

    @property
    def delete_mode(self) -> bool:
        return self._query_mode == QueryMode.Delete

    def table(self, val: str | Table | None) -> Self:
        """ Set the table name """
        self._select_from.from_expr.table = val
        return self

    def select(self, *vals: str | Col) -> Self:
        """ Set the selected cols """
        self._select_from.select.cols = vals
        return self

    def update(self, *assignations: Condition) -> Self:
        """ Update """
        self._update_set._set.assignations = assignations
        return self

    def distinct(self) -> Self:
        """ Filter the rows to remove duplicates (by selected columns)"""
        self._select_from.select.unique = True
        return self

    def head(self, length: int = 1) -> Self:
        """ Filter on the first N rows """
        self._limit = length
        return self

    def filter(self, condition: Condition) -> Self:
        """ Filter the rows according to the given condition """
        if any(x in condition.build() for x in set(e.value for e in AggFunction)):
            self._having.condition = condition
        else:
            self._where.condition = condition
        return self

    def group_by(self, *cols_names: str) -> Self:
        """ Group the rows of the specified columns """
        self._group_by.cols = cols_names
        return self

    def join(self, table: str | Table, condition: Condition | str | Iterable[str], join_type: JoinType | None = JoinType.Inner) -> Self:
        """ Join two tables together """
        self._joins.append(Join(table=table if isinstance(table, Table) else Table(name=table), condition=condition, join_type=join_type, compact=self._compact))
        return self

    def join_many(self, *joins_params: Tuple[str | Table, Condition | str | Iterable[str]] | Tuple[str | Table, Condition | str | Iterable[str], JoinType | None]) -> Self:
        """ Join more than two tables together """
        for join_params in joins_params:
            self.join(*join_params)
        return self

    def union(self, other_request: Self) -> Union:
        return Union(self, other_request, compact=self._compact)

    def __and__(self, other_request: Self) -> Union:
        return self.union(other_request)

    def __rand__(self, other_request: Self) -> Union:
        return self.__and__(other_request)



import re
from typing import Self

from typeguard import typechecked
from choc_expr import Expr as ChocExpr


@typechecked
class Table(ChocExpr):
    """ Represent a SQL table name """

    def __init__(
            self,
            name: str,
            alias: str | None = None,
            schema: str | None = None
    ) -> None:
        if not name:
            raise ValueError("The name parameter must not be empty")
        match = re.search(r"^([A-Za-z_\s]+\.)?([A-Za-z_\s()]+){1}((?:@|:)[A-Za-z_\s]*)?$", name)
        if not match:
            raise ValueError(f"Forbidden characters in expr '{name}'")
        r, n, a = match.groups()
        self._name = n
        self._alias = alias or (a[1:] if a else None)
        self._schema = schema or (r[:-1] if r else None)
        super().__init__("@{_schema}:{_schema}.:;{_name}@{_alias}: AS {_alias}:;")

    def alias(self, name: str) -> Self:
        """ Set an alias """
        self._alias = name
        return self

    @property
    def full_name(self) -> str:
        return str(self).split("AS")[0].strip()



from typeguard import typechecked
from choc_expr import Expr as ChocExpr


@typechecked
class Delete(ChocExpr):
    """ Delete expression """
    def __init__(
            self,
            compact: bool = True
    ) -> None:
        super().__init__("DELETE", compact=compact)



from .expr import *  # noqa
from .agg_function import AggFunction  # noqa
from .sql_function import SqlFunction  # noqa
from .operator import Operator  # noqa
from .ordering import Ordering  # noqa
from .sql_type import SqlType  # noqa
from .join_type import JoinType  # noqa
from .utils import quote_expr, to_bool  # noqa
from .shortcut import concat, count, sum, asc, desc, upper, lower, second, minute, hour, day, month, year  # noqa
from .query_mode import QueryMode  # noqa
from .alter_mode import AlterMode  # noqa
from .view_mode import ViewMode  # noqa


from __future__ import annotations
from typing import Self, TYPE_CHECKING

if TYPE_CHECKING:
    from .col import Col
    from .when import When

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from ..utils import quote_expr
from ..operator import Operator


@typechecked
class Condition(ChocExpr):
    """ SQL condition """
    def __init__(
            self,
            left_value: int | float | str | Col | Self,
            op: Operator,
            right_value: int | float | str | Col | When | Self,
            negate: bool = False
    ) -> None:
        self._left_value = left_value
        self._op = op
        self._right_value = right_value
        self._negate = negate

        super().__init__("@{negate}:NOT :;({left_val} {op} {right_val})")

    @property
    def negate(self):
        return self._negate

    @property
    def op(self):
        return self._op.value

    @property
    def left_val(self):
        return quote_expr(self._left_value)

    @property
    def right_val(self):
        try:
            return f"({self._right_value.build()})" if self._op == Operator.In else self._right_value.build()
        except AttributeError:
            pass
        return quote_expr(self._right_value)

    def __and__(self, other: int | float | str | Col | Self) -> Self:
        return Condition(left_value=self, op=Operator.And, right_value=other)

    def __rand__(self, other: int | float | str | Col | Self) -> Self:
        return self.__and__(other)

    def __or__(self, other: int | float | str | Col | Self) -> Self:
        return Condition(left_value=self, op=Operator.Or, right_value=other)

    def __ror__(self, other: int | float | str | Col | Self) -> Self:
        return self.__or__(other)

    def __invert__(self) -> Self:
        return Condition(left_value=self._left_value, op=self._op, right_value=self._right_value, negate=True)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .table import Table

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .from_expr import FromExpr
from .delete import Delete


@typechecked
class DeleteFrom(ChocExpr):
    """ Delete expression """
    def __init__(
            self,
            table: str | Table | None = None,
            compact: bool = True
    ) -> None:
        self._delete = Delete(compact=compact)
        self._from_expr = FromExpr(table=table, compact=compact)
        super().__init__("{delete~}{from_expr}", compact=compact)

    @property
    def delete(self):
        return self._delete

    @property
    def from_expr(self):
        return self._from_expr

    @property
    def buildable(self) -> bool:
        return self._from_expr.buildable



from typing import Iterable

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .condition import Condition
from ..utils import quote_expr


@typechecked
class When(ChocExpr):
    """ Conditional 'Case When' expression """
    def __init__(
            self,
            conditions: Iterable[Condition],
            returned_vals: Iterable[str | int | float],
            else_returned_val: str | int | float | None = None,
            compact: bool = False
    ):
        if len(conditions) != len(returned_vals):
            raise ValueError("'conditions' and 'returned_vals' must have the same length")

        if len(returned_vals) == 0:
            raise ValueError("CaseWhen must not be empty")

        self._items = [(condition, returned_val) for condition, returned_val in zip(conditions, returned_vals)]
        self._when_then_items = [
            f"WHEN {condition} THEN {quote_expr(returned_val)}" for condition, returned_val in zip(conditions, returned_vals)
        ]
        self._else = f"ELSE {quote_expr(else_returned_val)}" if else_returned_val else ""
        super().__init__("CASE\n{$(_when_then_items)~}{_else~}END", compact=compact, list_join_sep="\n")



from .table import Table  # noqa
from .from_expr import FromExpr  # noqa
from .select_from import SelectFrom  # noqa
from .select import Select  # noqa
from .limit import Limit  # noqa
from .query import Query  # noqa
from .col import Col  # noqa
from .col_when import ColWhen  # noqa
from .when import When  # noqa
from .condition import Condition  # noqa
from .where import Where  # noqa
from .having import Having  # noqa
from .group_by import GroupBy  # noqa
from .order_by import OrderBy  # noqa
from .join import Join  # noqa
from .set import Set  # noqa
from .update import Update  # noqa
from .update_set import UpdateSet  # noqa
from .delete import Delete  # noqa
from .delete_from import DeleteFrom  # noqa
from .union import Union  # noqa
from .view import View  # noqa



from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .condition import Condition


@typechecked
class Having(ChocExpr):
    """ Having expression """
    def __init__(
            self,
            condition: Condition | None = None,
            compact: bool = True
    ) -> None:
        self.condition = condition
        super().__init__("HAVING {condition}", compact=compact)

    @property
    def condition(self) -> Condition | None:
        return self._condition

    @condition.setter
    def condition(self, value: Condition | None) -> None:
        self._condition = value

    @property
    def buildable(self) -> bool:
        return self.condition is not None



from choc_expr import Expr as ChocExpr


class Union(ChocExpr):

    def __init__(self, left_query, right_query, compact: bool = False) -> None:
        self._left_query = left_query
        self._right_query = right_query
        super().__init__("{_left_query~}UNION~{_right_query}", compact=compact)



from __future__ import annotations
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from .table import Table
    from .col import Col

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .from_expr import FromExpr
from .select import Select


@typechecked
class SelectFrom(ChocExpr):
    """ Select expression """
    def __init__(
            self,
            table: str | Table | None = None,
            cols: Iterable[str | Col] | None = None,
            unique: bool = False,
            compact: bool = True
    ) -> None:
        self._select = Select(cols=cols, unique=unique, compact=compact)
        self._from_expr = FromExpr(table=table, compact=compact)
        super().__init__("{select~}{from_expr}", compact=compact)

    @property
    def select(self):
        return self._select

    @property
    def from_expr(self):
        return self._from_expr

    @property
    def buildable(self) -> bool:
        return self._from_expr.buildable



from typing import Iterable

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from ..join_type import JoinType
from .table import Table
from .condition import Condition


@typechecked
class Join(ChocExpr):
    """ Select expression """
    def __init__(
            self,
            table: Table,
            condition: Condition | str | Iterable[str],
            join_type: JoinType = JoinType.Inner,
            compact: bool = True
    ) -> None:
        if isinstance(condition, str):
            condition = [condition]
        self._table = table
        self._condition = condition
        self._join_type = join_type
        self._using = not isinstance(condition, Condition)
        super().__init__("{_join_type.value} JOIN {_table~}@{_using}:USING {$(_condition)}:ON {_condition};", compact=compact)



from chocolatine import Query, Col as _, month, year, sum, count, QueryMode, When, View, ViewMode
from chocolatine.sql_type import SqlType


def test_query_1a():
    """ How to display the first and last names of all actors from the table `actor` ? """
    assert str(Query(
        compact=False,
        table="actor",
        cols=("first_name", "last_name"),
    )) == """\
SELECT first_name, last_name
FROM actor
"""


def test_query_1b():
    """ How to display the first and last name of each actor in a single column in upper case letters (name the column `actor_name`) ? """
    assert str(Query(
        compact=False,
        table="actor",
        cols=((_("first_name") & " " & _("last_name")).upper().alias("actor_name"),)
    )) == """\
SELECT UPPER(CONCAT(first_name, ' ', last_name)) AS actor_name
FROM actor
"""


def test_query_2a():
    """ How to find the ID number, first name, and last name of an actor, of whom you know only the first name, "Joe." ? """
    assert str(Query(
        compact=False,
        table="actor",
        cols=("actor_id", "first_name", "last_name"),
        filters=(_("first_name") == "Joe",)
    )) == """\
SELECT actor_id, first_name, last_name
FROM actor
WHERE (first_name = 'Joe')
"""


def test_query_2b():
    """ How to find all actors whose last name contain the letters `GEN` ? """
    assert str(Query(
        compact=False,
        table="actor",
        filters=(_("last_name") >> r"%GEN%",)
    )) == """\
SELECT *
FROM actor
WHERE (last_name LIKE '%GEN%')
"""


def test_query_2c():
    """ How to find all actors whose last names contain the letters `LI` and ordering the rows by last name and first name ? """
    assert str(Query(
        compact=False,
        table="actor",
        cols=(">:last_name", ">:first_name",),
        filters=(_("last_name") >> r"%LI%",),
    )) == """\
SELECT last_name, first_name
FROM actor
WHERE (last_name LIKE '%LI%')
ORDER BY last_name ASC, first_name ASC
"""


def test_query_2d():
    """ How to, using `IN`, display the `country_id` and `country` columns of the following countries Afghanistan, Bangladesh, and China ? """
    assert str(Query(
        compact=False,
        table="country",
        cols=("country_id", "country"),
        filters=(_("country") << ("Afghanistan", "Bangladesh", "China"),),
    )) == """\
SELECT country_id, country
FROM country
WHERE (country IN ('Afghanistan', 'Bangladesh', 'China'))
"""


def test_query_3a():
    """ Add a `middle_name` column to the table `actor`. Position it between `first_name` and `last_name`. Hint: you will need to specify the data type. """
    assert str(Query.add_col(
        compact=False,
        table="actor",
        col=_("middle_name", type=SqlType.String),
        after=_("first_name")
    )) == """\
ALTER TABLE actor
ADD COLUMN middle_name VARCHAR(255) AFTER first_name
"""


def test_query_3b():
    """ You realize that some of these actors have tremendously long last names. Change the data type of the `middle_name` column to `blobs` """
    assert str(Query.rename_col(
        compact=False,
        table="actor",
        col="middle_name",
        new_col="blob"
    )) == """\
ALTER TABLE actor
RENAME COLUMN middle_name blob
"""


def test_query_3c():
    """ Now delete the `middle_name` column. """
    assert str(Query.drop_col(
        compact=False,
        table="actor",
        col="middle_name"
    )) == """\
ALTER TABLE actor
DROP COLUMN middle_name
"""


def test_query_4a():
    """ How to list the last names of actors, as well as how many actors have that last name ? """
    assert str(Query(
        compact=False,
        table="actor",
        cols=("last_name", count().alias("count")),
        groups=("last_name",),
    )) == """\
SELECT last_name, COUNT(*) AS count
FROM actor
GROUP BY last_name
"""


def test_query_4b():
    """ How to list last names of actors and the number of actors who have that last name, but only for names that are shared by at least two actors ? """
    assert str(Query(
        compact=False,
        table="actor",
        cols=("last_name", count().alias("count")),
        groups=("last_name",),
        filters=(count() > 1,),
    )) == """\
SELECT last_name, COUNT(*) AS count
FROM actor
GROUP BY last_name
HAVING (COUNT(*) > 1)
"""


def test_query_4c():
    """ Oh, no! The actor `HARPO WILLIAMS` was accidentally entered in the `actor` table as `GROUCHO WILLIAMS`, the name of Harpo"s second cousin"s husbands yoga teacher. Write a query to fix the record. """
    assert str(Query(
        compact=False,
        table="actor",
        query_mode=QueryMode.Update,
        assignations=((_("first_name") == "HARPO",)),
        filters=(((_("first_name") == "GROUCHO")) & (_("last_name") == "WILLIAMS"),)
    )) == """\
UPDATE actor
SET (first_name = 'HARPO')
WHERE ((first_name = 'GROUCHO') AND (last_name = 'WILLIAMS'))
"""


def test_query_4d():
    """ Perhaps we were too hasty in changing `GROUCHO` to `HARPO`. It turns out that `GROUCHO` was the correct name after all! In a
        single query, if the first name of the actor is currently `HARPO`, change it to `GROUCHO`. Otherwise, change the first name to
        `MUCHO GROUCHO`, as that is exactly what the actor will be with the grievous error. BE CAREFUL NOT TO CHANGE THE FIRST NAME OF
        EVERY ACTOR TO `MUCHO GROUCHO`, HOWEVER! (Hint: update the record using a unique identifier. """
    assert str(Query(
        compact=False,
        table="actor",
        query_mode=QueryMode.Update,
        assignations=(_("first_name") == When((_("first_name") == "HARPO",), ("GROUCHO",), "MUCHO GROUCHO"),),
        filters=((_("actor_id") == 172,))
    )) == """\
UPDATE actor
SET (first_name = CASE WHEN (first_name = 'HARPO') THEN 'GROUCHO' ELSE 'MUCHO GROUCHO' END)
WHERE (actor_id = 172)
"""


# 5a. You cannot locate the schema of the `address` table. Which query would you use to re-create it?

#   SHOW CREATE TABLE address;


def test_query_6a():
    """ How to, using `JOIN`, display the first and last names, as well as the address, of each staff member (use the tables `staff` and `address`) ? """
    assert str(Query(
        compact=False,
        table="staff",
        cols=("first_name", "last_name", "address"),
        joins=(("address", "address_id"),),
    )) == """\
SELECT first_name, last_name, address
FROM staff
INNER JOIN address
USING address_id
"""


def test_query_6b():
    """ How to, using `JOIN`, to display the total amount rung up by each staff member in August of 2005 (use tables `staff` and `payment`) ? """
    assert str(Query(
        compact=False,
        table="staff",
        cols=("first_name", "last_name", sum("amount")),
        joins=(("payment", "staff_id"),),
        filters=((month("payment_date") == 8) & (year("payment_date") == 2005),),
        groups=("staff_id",),
    )) == """\
SELECT first_name, last_name, SUM(amount)
FROM staff
INNER JOIN payment
USING staff_id
WHERE ((MONTH(payment_date) = 8) AND (YEAR(payment_date) = 2005))
GROUP BY staff_id
"""


def test_query_6c():
    """ How to list each film and the number of actors who are listed for that film (use tables `film_actor` and `film` and inner join) ? """
    assert str(Query(
        compact=False,
        table="film",
        cols=("title", count().alias("<:actor_count")),
        joins=(("film_actor", "film_id"),),
        groups=("title",),
    )) == """\
SELECT title, COUNT(*) AS actor_count
FROM film
INNER JOIN film_actor
USING film_id
GROUP BY title
ORDER BY actor_count DESC
"""


def test_query_6d():
    """ How many copies of the film `Hunchback Impossible` exist in the inventory system ? """
    assert str(Query(
        compact=False,
        table="film",
        cols=("title", count().alias("copies_count")),
        filters=(_("title") == "Hunchback Impossible",),
        joins=(("inventory", "film_id"),),
        groups=("title",)
    )) == """\
SELECT title, COUNT(*) AS copies_count
FROM film
INNER JOIN inventory
USING film_id
WHERE (title = 'Hunchback Impossible')
GROUP BY title
"""


def test_query_6e():
    """ How to, using the tables `payment` and `customer` and the `JOIN` command, list the total paid by each customer ? """
    assert str(Query(
        compact=False,
        table="customer",
        cols=(">:last_name", "first_name", sum("amount").alias("total_paid_amount")),
        joins=(("payment", "customer_id"),),
        groups=("last_name",)
    )) == """\
SELECT last_name, first_name, SUM(amount) AS total_paid_amount
FROM customer
INNER JOIN payment
USING customer_id
GROUP BY last_name
ORDER BY last_name ASC
"""


def test_query_7a():
    """ How to, using subqueries, display the titles of movies starting with the letters `K` and `Q` whose language is English ? """
    assert str(Query(
        compact=False,
        table="film",
        cols=("title",),
        filters=(
            ((_("title") >> "K%") | (_("title") >> "Q%")) & _("language_id") << Query(
                table="language",
                cols=("language_id",),
                filters=(_("name") == "English",),
            ),
        )
    )) == """\
SELECT title
FROM film
WHERE (((title LIKE 'K%') OR (title LIKE 'Q%')) AND (language_id IN (SELECT language_id FROM language WHERE (name = 'English'))))
"""


def test_query_7b():
    """ How to, using subqueries, display all actors who appear in the film `Alone Trip ? """
    assert str(Query(
        compact=False,
        table="actor",
        cols=("first_name", "last_name"),
        filters=(
            _("actor_id") << Query(
                table="film_actor",
                cols=("actor_id",),
                filters=(
                    _("film_id") << Query(
                        table="film",
                        cols=("film_id",),
                        filters=(_("title") == "Alone Trip",)
                    ),
                ),
            ),
        )
    )) == """\
SELECT first_name, last_name
FROM actor
WHERE (actor_id IN (SELECT actor_id FROM film_actor WHERE (film_id IN (SELECT film_id FROM film WHERE (title = 'Alone Trip')))))
"""


def test_query_7c():
    """ How to, knowing you want to run an email marketing campaign in Canada, for which you will need the names and email addresses of all Canadian customers, retrieve this information (use joins) ? """
    assert str(Query(
        compact=False,
        cols=("first_name", "last_name", "email"),
        table="customer",
        joins=(("address", "address_id"), ("city", "city_id"), ("country", "country_id")),
        filters=(_("country") == "canada",)
    )) == """\
SELECT first_name, last_name, email
FROM customer
INNER JOIN address
USING address_id
INNER JOIN city
USING city_id
INNER JOIN country
USING country_id
WHERE (country = 'canada')
"""


def test_query_7d():
    """ How to identify all movies categorized as family films ? """
    assert str(Query(
        compact=False,
        table="film",
        cols=("title", "name"),
        joins=(("film_category", "film_id"), ("category", "category_id")),
        filters=(_("name") == "family",),
    )) == """\
SELECT title, name
FROM film
INNER JOIN film_category
USING film_id
INNER JOIN category
USING category_id
WHERE (name = 'family')
"""


def test_query_7e():
    """ How to display the most frequently rented movies in descending order ? """
    assert str(Query(
        compact=False,
        table="film",
        cols=("title", count().alias("<:rentals")),
        joins=(("inventory", "film_id"), ("rental", "inventory_id")),
        groups=("title",)
    )) == """\
SELECT title, COUNT(*) AS rentals
FROM film
INNER JOIN inventory
USING film_id
INNER JOIN rental
USING inventory_id
GROUP BY title
ORDER BY rentals DESC
"""


def test_query_7f():
    """ How much business, in dollars, each store brought in ? """
    assert str(Query(
        compact=False,
        table="payment",
        cols=("store_id", sum("amount").alias("revenue")),
        joins=(("rental", "rental_id"), ("inventory", "inventory_id"), ("store", "store_id")),
        groups=("store_id",)
    )) == """\
SELECT store_id, SUM(amount) AS revenue
FROM payment
INNER JOIN rental
USING rental_id
INNER JOIN inventory
USING inventory_id
INNER JOIN store
USING store_id
GROUP BY store_id
"""


def test_query_7g():
    """ How to display for each store its store ID, city, and country ? """
    assert str(Query(
        compact=False,
        table="store",
        cols=("store_id", "city", "country"),
        joins=(("address", "address_id"), ("city", "city_id"), ("country", "country_id")),
    )) == """\
SELECT store_id, city, country
FROM store
INNER JOIN address
USING address_id
INNER JOIN city
USING city_id
INNER JOIN country
USING country_id
"""


def test_query_7h():
    """ List the top five genres in gross revenue in descending order (use the tables category, film_category, inventory, payment, and rental) ? """
    assert str(Query(
        compact=False,
        table="payment",
        cols=(sum("amount").alias("<:total_amount"), "name"),
        joins=(("rental", "rental_id"), ("inventory", "inventory_id"), ("category", "category_id")),
        groups=("name",)
    )) == """\
SELECT SUM(amount) AS total_amount, name
FROM payment
INNER JOIN rental
USING rental_id
INNER JOIN inventory
USING inventory_id
INNER JOIN category
USING category_id
GROUP BY name
ORDER BY total_amount DESC
"""


def test_query_8a():
    """ In your new role as an executive, you would like to have an easy way of viewing the Top five genres by gross revenue. Use the
        solution from the problem above to create a view. If you havent solved 7h, you can substitute another query to create a view. """
    assert str(View(
        name="top_five_genres",
        compact=False,
        query=Query(
            compact=False,
            table="payment",
            cols=(sum("amount").alias("<:total_sales"), "name:genre"),
            joins=(("rental", "rental_id"), ("inventory", "inventory_id"), ("film_category", "film_id"), ("category", "category_id")),
            groups=("name",),
            limit=5
        ))) == """\
CREATE VIEW top_five_genres AS
SELECT SUM(amount) AS total_sales, name AS genre
FROM payment
INNER JOIN rental
USING rental_id
INNER JOIN inventory
USING inventory_id
INNER JOIN film_category
USING film_id
INNER JOIN category
USING category_id
GROUP BY name
ORDER BY total_sales DESC
LIMIT 5
"""


def test_query_8b():
    """ How would you display the view that you created in 8a? """
    assert str(Query(
        compact=False,
        table="top_five_genres"
    )) == """\
SELECT *
FROM top_five_genres
"""


def test_query_8c():
    """ How would you display the view that you created in 8a? """
    assert str(View(
        name="top_five_genres",
        mode=ViewMode.Drop,
        compact=False,
    )) == """\
DROP VIEW top_five_genres
"""



from typing import Iterable

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .col import Col


@typechecked
class Select(ChocExpr):
    """ Select expression """
    def __init__(
            self,
            cols: Iterable[Col | str] | None = None,
            unique: bool = False,
            compact: bool = True
    ) -> None:
        self.cols = cols
        self.unique = unique
        super().__init__("SELECT @{unique}:DISTINCT(:;@{empty}:*:{$(cols)};@{unique}:):;", compact=compact)

    @property
    def unique(self):
        return self._unique

    @unique.setter
    def unique(self, value):
        self._unique = value

    @property
    def empty(self):
        return len(self._cols) == 0

    @property
    def cols(self):
        return self._cols

    @cols.setter
    def cols(self, value):
        self._cols = [Col(col) if type(col) is str else col for col in (value or [])]



from typing import Iterable

from typeguard import typechecked
from choc_expr import Expr as ChocExpr

from .col import Col


@typechecked
class GroupBy(ChocExpr):
    """ GroupBy expression """
    def __init__(
            self,
            cols: Iterable[Col | str] | None = None,
            compact: bool = True
    ) -> None:
        self.cols = cols
        super().__init__("GROUP BY {$(cols)}", compact=compact)

    @property
    def cols(self):
        return self._cols

    @cols.setter
    def cols(self, value: Iterable[Col | str] | None):
        if value is not None and len(value) < 1:
            raise ValueError("The number of cols must be at least 1")
        self._cols = [Col(col) if type(col) is str else col for col in (value or [])]

    @property
    def buildable(self):
        return self.cols is not None and len(self.cols) > 0
